//
// Copyright (C) 2000-2003 Satoshi Imai
// Portions Copyright (C) 2017 David Bodenstab
//

#include "cwind.h"
#include <ctype.h>

//
// Cwind
//

static struct ProcArg {
	char *winText;		// (I)
	HWND hTarget;		// (O)
	Tcl_Interp *interp;	// (I)
	BOOL fExact;		// (I)
	BOOL fChild;		// (I)
} ProcArg;

static DWORD sleeptime = 0;

// Which mouse button?
enum mbutton { left=0, middle, right };
enum maction { down=0, up };

static int downup[][2] ={
  {MOUSEEVENTF_LEFTDOWN, MOUSEEVENTF_LEFTUP},
  {MOUSEEVENTF_MIDDLEDOWN, MOUSEEVENTF_MIDDLEUP},
  {MOUSEEVENTF_RIGHTDOWN, MOUSEEVENTF_RIGHTUP},
};

// Simple, inline math functions
static inline int
Abs( int value )
{
	return (value < 0)? -value : value;
}
static inline int
Max( int value1, int value2 )
{
	return (value1 > value2)? value1 : value2;
}


//
// Find a window
//
BOOL CALLBACK
EnumFindProc(
	HWND hWnd,
	LPARAM lParam
)
{
	char *text;
	int len;
	Tcl_DString dString;

	if ((len = GetWindowTextLength(hWnd)) < 1) {
		return TRUE;
	}
	if (!IsWindowVisible(hWnd)) {
		return TRUE;
	}
//	if (GetWindow(hWnd, GW_OWNER)) {
//		return TRUE;
//	}
	text = Tcl_Alloc(len+1);
	if (text == NULL) {
		return TRUE;
	}
	GetWindowText(hWnd, text, len+1);
	Tcl_ExternalToUtfDString(NULL, text, len, &dString);

	if (((struct ProcArg *)lParam)->fExact == TRUE) {
		if (!strcmp(Tcl_DStringValue(&dString), ((struct ProcArg *)lParam)->winText)) {
			((struct ProcArg *)lParam)->hTarget = hWnd;
			goto found;
		}
	}
	else {
		if (Tcl_StringMatch(Tcl_DStringValue(&dString), ((struct ProcArg *)lParam)->winText)) {
			((struct ProcArg *)lParam)->hTarget = hWnd;
			goto found;
		}
	}

	Tcl_DStringFree(&dString);
	Tcl_Free(text);
	return TRUE;

found:
	Tcl_DStringFree(&dString);
	Tcl_Free(text);
	return FALSE;
}

//
// List windows
//
BOOL CALLBACK
EnumListProc(
	HWND hWnd,
	LPARAM lParam
)
{
	char *text;
	int len;
	Tcl_DString dString;

	if ((len = GetWindowTextLength(hWnd)) < 1) {
		return TRUE;
	}
	if (!IsWindowVisible(hWnd)) {
		return TRUE;
	}
//	if (GetWindow(hWnd, GW_OWNER)) {
//		return TRUE;
//	}
	if (((struct ProcArg *)lParam)->fChild == TRUE) {
		HMODULE hmodule = GetModuleHandle("USER32.DLL");
		typedef int (FAR PASCAL *WINPROC) (HWND hwnd, PWINDOWINFO pwi);
		WINPROC proc = (WINPROC)GetProcAddress(hmodule, "GetWindowInfo");
		if (proc) {
			WINDOWINFO wi;
			if (proc(hWnd, &wi) && !(wi.dwExStyle & WS_EX_MDICHILD)) {
				return TRUE;
			}
		}
	}
	text = Tcl_Alloc(len+1);
	if (text == NULL) {
		return TRUE;
	}
	GetWindowText(hWnd, text, len+1);
	Tcl_ExternalToUtfDString(NULL, text, len, &dString);

	Tcl_AppendElement(((struct ProcArg *)lParam)->interp, Tcl_DStringValue(&dString));

	Tcl_DStringFree(&dString);
	Tcl_Free(text);
	return TRUE;
}

//
// Helper routine to parse ?-exact? window-title arguments
// Sets ProcArg values, and optional child pattern Tcl_Obj
// Returns TCL_OK or TCL_ERROR
//
static int
exactWindowTitleArgs( Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], Tcl_Obj **child )
{
	int len;
	Tcl_Obj *pattern, *parent;

	if ((objc != 2) && (objc != 3)) {
		Tcl_WrongNumArgs(interp, 1, objv, "?-exact? pattern");
		return TCL_ERROR;
	}

	ProcArg.hTarget = NULL;

	if ( objc == 2 ) {
		ProcArg.fExact = FALSE;
	  	pattern = objv[1];
	} else {
	  	// objc == 3
		char *exact = Tcl_GetString(objv[1]);
		// I see what the author intended: match a partial string but
		// what if the string is just "-", or even, ""?
		if (!strncmp(exact, "-exact", strlen(exact)))
			ProcArg.fExact = TRUE;
		else {
			Tcl_WrongNumArgs(interp, 1, objv, "?-exact? pattern");
			return TCL_ERROR;
		}
	  	pattern = objv[2];
	}

	// Deal with the window title argument
	if ( Tcl_ListObjLength(interp,pattern,&len) != TCL_OK )
	        return TCL_ERROR;
	if (len != 1 && len != 2) {
		Tcl_AppendResult(interp, "Window title must be a one- or two-element list.", NULL);
		return TCL_ERROR;
	}
	if ( Tcl_ListObjIndex(interp,pattern,0,&parent) != TCL_OK )
		return TCL_ERROR;
	ProcArg.winText = Tcl_GetString(parent);

	if ( len == 2 ) {
		if ( Tcl_ListObjIndex(interp,pattern,1,child) != TCL_OK )
			return TCL_ERROR;
	} else
	  	*child = NULL;

	return TCL_OK;
}

//
// GetHwnd title
//
static int
Cwind_GetTextObjCmd(
	ClientData clientData,
	Tcl_Interp *interp,
	int objc,
	Tcl_Obj *CONST objv[]
)
{
	char *text;
	int len;
	Tcl_DString dString;

	if ((len = GetWindowTextLength(GetForegroundWindow())) < 0) {
		Tcl_AppendResult(interp, "WindowText is empty.", NULL);
		return TCL_ERROR;
	}
	text = Tcl_Alloc(len+1);
	if (text == NULL) {
		return TCL_ERROR;
	}
	GetWindowText(GetForegroundWindow(), text, len+1);
	Tcl_ExternalToUtfDString(NULL, text, len, &dString);
	
	Tcl_AppendResult(interp, Tcl_DStringValue(&dString), NULL);
	Tcl_DStringFree(&dString);
	Tcl_Free(text);
	return TCL_OK;
}

//
// Top, Bottom, Topmost, Notopmost
//
static int
windowpos( Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], HWND position )
{
  	char *arg2;

	if ((objc != 2) && (objc != 3)) {
		Tcl_WrongNumArgs(interp, 1, objv, "?-exact? pattern");
		return TCL_ERROR;
	}
	ProcArg.hTarget = NULL;
	ProcArg.winText = arg2 = Tcl_GetString(objv[1]);
	ProcArg.fExact = FALSE;

	if (objc == 3) {
		if (!strncmp(arg2, "-exact", strlen(arg2))) {
			ProcArg.fExact = TRUE;
			ProcArg.winText = Tcl_GetString(objv[2]);
		}
		else {
			Tcl_WrongNumArgs(interp, 1, objv, "?-exact? pattern");
			return TCL_ERROR;
		}
	}

	EnumWindows(EnumFindProc, (LPARAM)&ProcArg);

	if (ProcArg.hTarget == NULL) {
		Tcl_AppendResult(interp, "No such window exists.", NULL);
		return TCL_ERROR;
	}

	SetWindowPos(ProcArg.hTarget, position, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
	return TCL_OK;
}

static int
Cwind_BottomObjCmd( ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[] )
{
	return windowpos( interp, objc, objv, HWND_BOTTOM );
}

static int
Cwind_TopObjCmd( ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[] )
{
	return windowpos( interp, objc, objv, HWND_TOP );
}

static int
Cwind_TopMostObjCmd( ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[] )
{
	return windowpos( interp, objc, objv, HWND_TOPMOST );
}

static int
Cwind_NoTopMostObjCmd( ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[] )
{
	return windowpos( interp, objc, objv, HWND_NOTOPMOST );
}

//
// Show ?-exact? window-title
//
static int
Cwind_ShowObjCmd(
	ClientData clientData,
	Tcl_Interp *interp,
	int objc,
	Tcl_Obj *CONST objv[]
)
{
	HWND hParent;
	Tcl_Obj *MDIchild;

	if ( exactWindowTitleArgs( interp, objc, objv, &MDIchild ) != TCL_OK )
		return TCL_ERROR;

	EnumWindows(EnumFindProc, (LPARAM)&ProcArg);

	if (ProcArg.hTarget == NULL) {
		Tcl_AppendResult(interp, "No such window exists.", NULL);
		return TCL_ERROR;
	}

	hParent = ProcArg.hTarget;

	if (MDIchild) {
		ProcArg.hTarget = NULL;
		ProcArg.winText = Tcl_GetString(MDIchild);
		EnumChildWindows(hParent, EnumFindProc, (LPARAM)&ProcArg);
		if (ProcArg.hTarget == NULL) {
			Tcl_AppendResult(interp, "No such child window exists.", NULL);
			return TCL_ERROR;
		}
	}

	if (!MySetForegroundWindow(hParent)) {
		ShowWindow(hParent, SW_MINIMIZE);
		ShowWindow(hParent, SW_RESTORE);
	}

	if (MDIchild)
		BringWindowToTop(ProcArg.hTarget);

	if (IsIconic(ProcArg.hTarget))
		ShowWindow(ProcArg.hTarget, SW_RESTORE);

	return TCL_OK;
}

//
// Lookup special character names
//
BOOL
ScLookup(
	char *name,	// (I)
	BYTE *vkey,	// (O)
	DWORD *isup	// (O)
) {
	struct SpecialChar *psc;
	int i;

	psc = &SpecialChar[0];
	for (i = 0; psc[i].sc_name; i++) {
		if ((psc[i].sc_name[0] == name[0]) &&
				!stricmp(psc[i].sc_name, name)) {
			*vkey = psc[i].sc_vkey;
			*isup = psc[i].sc_flag;
			return TRUE;
		}
	}
	return FALSE;
}

//
// Windows commands
// 
UINT
WmLookup(
	Tcl_Interp *interp,
	char *str
) {
	struct WmCommand *wmc;
	UINT cmd;
	int i;

	if (Tcl_GetInt(interp, str, (int *)&cmd) == TCL_OK) {
		return cmd;
	}
	Tcl_ResetResult(interp);
	wmc = &WmCommand[0];
	for (i = 0; wmc[i].str; i++) {
		if ((wmc[i].str[0] == str[0]) && !stricmp(wmc[i].str, str)) {
			return wmc[i].cmd;
		}
	}
	return 0;
}

//
// Send
//
static int
Cwind_SendObjCmd(
	ClientData clientData,
	Tcl_Interp *interp,
	int objc,
	Tcl_Obj *CONST objv[]
)
{
	char *mesg;
	BYTE vkey;
	DWORD isup;
	BOOL iscap;
	SHORT code;
	int i;

	if (objc < 2) {
		Tcl_WrongNumArgs(interp, 1, objv, "string ...");
		return TCL_ERROR;
	}

	for (i = 1; i < objc; i++) {
		mesg = Tcl_GetString(objv[i]);
		if (mesg == NULL) {
			Tcl_AppendResult(interp, "No message exists.", NULL);
			return TCL_ERROR;
		}
		if (ScLookup(mesg, &vkey, &isup)) {
			if (sleeptime) Sleep(sleeptime);
			keybd_event(vkey, 0, isup, 0);
			if ((vkey == VK_MENU) || (vkey == VK_CONTROL) || (vkey == VK_SHIFT))
				continue;
			keybd_event(vkey, 0, KEYEVENTF_KEYUP, 0);
		}
		else {
			Tcl_UniChar *us=Tcl_GetUnicode(objv[i]);
			for ( ; *us; us++) {
				if (sleeptime) Sleep(sleeptime);
				if (*us>255) {
					Tcl_AppendResult(interp, "can't send Unicode character.", NULL);
					return TCL_ERROR;
				}
				code = VkKeyScan((char)*us);
				vkey = LOBYTE(code);
				iscap = HIBYTE(code);
				if (iscap&1) keybd_event(VK_SHIFT, 0, 0, 0);
				if (iscap&2) keybd_event(VK_CONTROL, 0, 0, 0);
				if (iscap&4) keybd_event(VK_MENU, 0, 0, 0);
				keybd_event(vkey, 0, 0, 0);
				keybd_event(vkey, 0, KEYEVENTF_KEYUP, 0);
				if (iscap&4) keybd_event(VK_MENU, 0, KEYEVENTF_KEYUP, 0);
				if (iscap&2) keybd_event(VK_CONTROL, 0, KEYEVENTF_KEYUP, 0);
				if (iscap&1) keybd_event(VK_SHIFT, 0, KEYEVENTF_KEYUP, 0);
			}
		}
	}
	return TCL_OK;
}

//
// Getcb
//
static int
Cwind_GetCbObjCmd(
	ClientData clientData,
	Tcl_Interp *interp,
	int objc,
	Tcl_Obj *CONST objv[]
)
{
	if (!OpenClipboard(NULL)) {
		Tcl_AppendResult(interp, "Cannot open clipboard.", NULL);
		return TCL_ERROR;
	}
	if (!IsClipboardFormatAvailable(CF_TEXT)) {
		CloseClipboard();
		Tcl_AppendResult(interp, "Clipboard is empty.", NULL);
		return TCL_ERROR;
	}
	HANDLE hClip = GetClipboardData(CF_TEXT);
	if (!hClip) {
		CloseClipboard();
		Tcl_AppendResult(interp, "Cannot get clipboard data.", NULL);
		return TCL_ERROR;
	}
	LPVOID pClip = GlobalLock(hClip);
	if (!pClip) {
		CloseClipboard();
		Tcl_AppendResult(interp, "Cannot lock clipboard.", NULL);
		return TCL_ERROR;
	}
	DWORD len = strlen((char *)pClip);
	if (len == 0) {
		GlobalUnlock(hClip);
		CloseClipboard();
		Tcl_AppendResult(interp, "Clipboard data size is zero.", NULL);
		return TCL_ERROR;
	}
	char *pbuf = Tcl_Alloc(len+1);
	if (pbuf == NULL) {
		GlobalUnlock(hClip);
		CloseClipboard();
		Tcl_AppendResult(interp, "Not enough memory.", NULL);
		return TCL_ERROR;
	}
	memcpy(pbuf, pClip, len);
	pbuf[len] = '\0';

	Tcl_DString dString;
	Tcl_ExternalToUtfDString(NULL, pbuf, len, &dString);

	Tcl_AppendResult(interp, Tcl_DStringValue(&dString), NULL);

	GlobalUnlock(hClip);
	Tcl_Free(pbuf);
	Tcl_DStringFree(&dString);
	CloseClipboard();
	return TCL_OK;
}

//
// Putcb
//
static int
Cwind_PutCbObjCmd(
	ClientData clientData,
	Tcl_Interp *interp,
	int objc,
	Tcl_Obj *CONST objv[]
)
{
	if (objc != 2) {
		Tcl_WrongNumArgs(interp, 1, objv, "string");
		return TCL_ERROR;
	}
	if (!OpenClipboard(NULL)) {
		Tcl_AppendResult(interp, "Cannot open clipboard.", NULL);
		return TCL_ERROR;
	}

	DWORD len = strlen(Tcl_GetString(objv[1]));
	Tcl_DString dString;
	Tcl_UtfToExternalDString(NULL, Tcl_GetString(objv[1]), len, &dString);
	len = strlen(Tcl_DStringValue(&dString));

	HGLOBAL hClip = GlobalAlloc(GMEM_MOVEABLE | GMEM_DDESHARE, len+1);
	if (!hClip) {
		CloseClipboard();
		Tcl_AppendResult(interp, "Not enough memory.", NULL);
		return TCL_ERROR;
	}
	LPVOID pClip = GlobalLock(hClip);
	if (!pClip) {
		CloseClipboard();
		Tcl_AppendResult(interp, "Cannot lock memory.", NULL);
		return TCL_ERROR;
	}
	memcpy(pClip, Tcl_DStringValue(&dString), len);
	((char *)pClip)[len] = '\0';

	Tcl_DStringFree(&dString);
	GlobalUnlock(hClip);

	EmptyClipboard();
	if (!SetClipboardData(CF_TEXT, hClip)) {
		CloseClipboard();
		Tcl_AppendResult(interp, "Cannot set clipboard data.", NULL);
		return TCL_ERROR;
	}
	CloseClipboard();

	return TCL_OK;
}

//
// List
//
static int
Cwind_WlistObjCmd(
	ClientData clientData,
	Tcl_Interp *interp,
	int objc,
	Tcl_Obj *CONST objv[]
)
{
	if (objc == 1) {
		ProcArg.interp = interp;
		ProcArg.fChild = FALSE;

		EnumWindows(EnumListProc, (LPARAM)&ProcArg);

		return TCL_OK;
	}

	if ((objc != 2) && (objc != 3)) {
		Tcl_WrongNumArgs(interp, 1, objv, "?-exact? pattern");
		return TCL_ERROR;
	}
	ProcArg.hTarget = NULL;
	ProcArg.winText = Tcl_GetString(objv[1]);
	ProcArg.fExact = FALSE;

	if (objc == 3) {
		if (!strncmp(Tcl_GetString(objv[1]), "-exact", strlen(Tcl_GetString(objv[1])))) {
			ProcArg.fExact = TRUE;
			ProcArg.winText = Tcl_GetString(objv[2]);
		}
		else {
			Tcl_WrongNumArgs(interp, 1, objv, "?-exact? pattern");
			return TCL_ERROR;
		}
	}

	EnumWindows(EnumFindProc, (LPARAM)&ProcArg);

	if (ProcArg.hTarget == NULL) {
		Tcl_AppendResult(interp, "No such window exists.", NULL);
		return TCL_ERROR;
	}

	ProcArg.interp	= interp;
	ProcArg.fChild = TRUE;

	EnumChildWindows(ProcArg.hTarget, EnumListProc, (LPARAM)&ProcArg);

	return TCL_OK;
}

//
// Sleeptime
//
static int
Cwind_SleeptimeObjCmd(
	ClientData clientData,
	Tcl_Interp *interp,
	int objc,
	Tcl_Obj *CONST objv[]
)
{
	int num;

	// Return the current value of sleeptime
	if (objc == 1)
	  	Tcl_SetObjResult( interp, Tcl_NewIntObj(sleeptime) );
	else {
		// Set a new sleeptime value
		if ( Tcl_GetIntFromObj(interp,objv[1],&num) != TCL_OK )
			return TCL_ERROR;
		sleeptime = num;
	}

	return TCL_OK;
}

//
// Restore
//
static int
Cwind_RestoreObjCmd(
	ClientData clientData,
	Tcl_Interp *interp,
	int objc,
	Tcl_Obj *CONST objv[]
)
{
	HWND hWnd = GetForegroundWindow();
	ShowWindow(hWnd, SW_RESTORE);
//	UpdateWindow(hWnd);

	return TCL_OK;
}

//
// Minimize
//
static int
Cwind_MinimizeObjCmd(
	ClientData clientData,
	Tcl_Interp *interp,
	int objc,
	Tcl_Obj *CONST objv[]
)
{
	HWND hWnd = GetForegroundWindow();
	ShowWindow(hWnd, SW_MINIMIZE);
//	UpdateWindow(hWnd);

	return TCL_OK;
}

//
// Maximize
//
static int
Cwind_MaximizeObjCmd(
	ClientData clientData,
	Tcl_Interp *interp,
	int objc,
	Tcl_Obj *CONST objv[]
)
{
	HWND hWnd = GetForegroundWindow();
	ShowWindow(hWnd, SW_MAXIMIZE);
//	UpdateWindow(hWnd);

	return TCL_OK;
}

//
// State
//
static int
Cwind_StateObjCmd(
	ClientData clientData,
	Tcl_Interp *interp,
	int objc,
	Tcl_Obj *CONST objv[]
)
{
	Tcl_Obj *MDIchild;

	if ( exactWindowTitleArgs( interp, objc, objv, &MDIchild ) != TCL_OK )
		return TCL_ERROR;

	EnumWindows(EnumFindProc, (LPARAM)&ProcArg);

	if (ProcArg.hTarget == NULL) {
		Tcl_AppendResult(interp, "No such window exists.", NULL);
		return TCL_ERROR;
	}

	if (MDIchild) {
		HWND hParent = ProcArg.hTarget;
		ProcArg.hTarget = NULL;
		ProcArg.winText = Tcl_GetString(MDIchild);
		EnumChildWindows(hParent, EnumFindProc, (LPARAM)&ProcArg);
		if (ProcArg.hTarget == NULL) {
			Tcl_AppendResult(interp, "No such child window exists.", NULL);
			return TCL_ERROR;
		}
	}

	if (IsIconic(ProcArg.hTarget)) {
		Tcl_AppendResult(interp, "Minimized", NULL);
	}
	else if (IsZoomed(ProcArg.hTarget)) {
		Tcl_AppendResult(interp, "Maximized", NULL);
	}
	else {
		Tcl_AppendResult(interp, "Normal", NULL);
	}

	return TCL_OK;
}

//
// Sendmessage
//
static int
Cwind_SendmessageObjCmd(
	ClientData clientData,
	Tcl_Interp *interp,
	int objc,
	Tcl_Obj *CONST objv[]
)
{
	HWND hWnd;
	UINT msg;
	WPARAM wparam=0;
	LPARAM lparam=0;

	if ((objc < 2) || (objc > 5)) {
		Tcl_WrongNumArgs(interp, 1, objv, "message ?wparam? ?lparam?");
		return TCL_ERROR;
	}
	msg = WmLookup(interp, Tcl_GetString(objv[1]));
	if (!msg) {
		Tcl_AppendResult(interp, "invalid message.", NULL);
		return TCL_ERROR;
	}
	if (objc > 2) {
		if (Tcl_GetInt(interp, Tcl_GetString(objv[2]), (int *)&wparam) != TCL_OK)
			return TCL_ERROR;
	}
	if (objc > 3) {
		if (Tcl_GetInt(interp, Tcl_GetString(objv[3]), (int *)&lparam) != TCL_OK)
			return TCL_ERROR;
	}
	hWnd = GetForegroundWindow();
	SendMessage(hWnd, msg, wparam, lparam);
	return TCL_OK;
}

//
// Postmessage
//
static int
Cwind_PostmessageObjCmd(
	ClientData clientData,
	Tcl_Interp *interp,
	int objc,
	Tcl_Obj *CONST objv[]
)
{
	HWND hWnd;
	UINT msg;
	WPARAM wparam=0;
	LPARAM lparam=0;

	if ((objc < 2) || (objc > 5)) {
		Tcl_WrongNumArgs(interp, 1, objv, "message ?wparam? ?lparam?");
		return TCL_ERROR;
	}
	msg = WmLookup(interp, Tcl_GetString(objv[1]));
	if (!msg) {
		Tcl_AppendResult(interp, "invalid message.", NULL);
		return TCL_ERROR;
	}
	if (objc > 2) {
		if (Tcl_GetInt(interp, Tcl_GetString(objv[2]), (int *)&wparam) != TCL_OK)
			return TCL_ERROR;
	}
	if (objc > 3) {
		if (Tcl_GetInt(interp, Tcl_GetString(objv[3]), (int *)&lparam) != TCL_OK)
			return TCL_ERROR;
	}
	hWnd = GetForegroundWindow();
	PostMessage(hWnd, msg, wparam, lparam);
	return TCL_OK;
}

//
// Getpos - Return a list containing the current mouse coordinates on
//          the screen.
//
static int
Cwind_GetposObjCmd(
	ClientData clientData,
	Tcl_Interp *interp,
	int objc,
	Tcl_Obj *CONST objv[]
)
{
	POINT point;
	Tcl_Obj *ov[2];
	
	if (GetCursorPos(&point) == 0) {
		Tcl_AppendResult(interp, "Cannot get cursor position.", NULL);
		return TCL_ERROR;
	}

	ov[0] = Tcl_NewIntObj(point.x);
	ov[1] = Tcl_NewIntObj(point.y);
	Tcl_SetObjResult(interp, Tcl_NewListObj(2,ov) );

	return TCL_OK;
}

//
// Getwpos - Return a list containing the current mouse coordinates on
//           the screen relative to the active window.
//
static int
Cwind_GetwposObjCmd(
	ClientData clientData,
	Tcl_Interp *interp,
	int objc,
	Tcl_Obj *CONST objv[]
)
{
	POINT point;
	Tcl_Obj *ov[2];
	
	if (GetCursorPos(&point) == 0) {
		Tcl_AppendResult(interp, "Cannot get cursor position.", NULL);
		return TCL_ERROR;
	}
	if (ScreenToClient(GetForegroundWindow(), &point) == 0) {
		Tcl_AppendResult(interp, "Cannot convert from screen to client.", NULL);
		return TCL_ERROR;
	}

	ov[0] = Tcl_NewIntObj(point.x);
	ov[1] = Tcl_NewIntObj(point.y);
	Tcl_SetObjResult(interp, Tcl_NewListObj(2,ov) );

	return TCL_OK;
}

//
// Setpos {x y}
// Set mouse position to {x y}
//
static int
Cwind_SetposObjCmd(
	ClientData clientData,
	Tcl_Interp *interp,
	int objc,
	Tcl_Obj *CONST objv[]
)
{
	POINT point;
	int len;
	Tcl_Obj *x, *y;

	if (objc != 2) {
		Tcl_WrongNumArgs(interp, 1, objv, "{x y}");
		return TCL_ERROR;
	}
	if ( Tcl_ListObjLength(interp,objv[1],&len) != TCL_OK )
	        return TCL_ERROR;
	if (len != 2) {
		Tcl_AppendResult(interp, "Cursor must be a two-element list.", NULL);
		return TCL_ERROR;
	}
	if ( Tcl_ListObjIndex(interp,objv[1],0,&x) != TCL_OK )
	        return TCL_ERROR;
	if ( Tcl_ListObjIndex(interp,objv[1],1,&y) != TCL_OK )
	        return TCL_ERROR;
	if ( Tcl_GetIntFromObj(interp,x,&point.x) != TCL_OK )
	        return TCL_ERROR;
	if ( Tcl_GetIntFromObj(interp,y,&point.y) != TCL_OK )
	        return TCL_ERROR;

	if (SetCursorPos(point.x, point.y) == 0) {
		Tcl_AppendResult(interp, "Cannot set cursor position.", NULL);
		return TCL_ERROR;
	}

	return TCL_OK;
}

//
// Setwpos {x y}
// Set the mouse position in the active window relative to the origin {0 0}
// at the upper left corner to coordinates (a two-element list).
//
static int
Cwind_SetwposObjCmd(
	ClientData clientData,
	Tcl_Interp *interp,
	int objc,
	Tcl_Obj *CONST objv[]
)
{
	POINT point;
	int len;
	Tcl_Obj *x, *y;
	
	if (objc != 2) {
		Tcl_WrongNumArgs(interp, 1, objv, "{x y}");
		return TCL_ERROR;
	}
	if ( Tcl_ListObjLength(interp,objv[1],&len) != TCL_OK )
	        return TCL_ERROR;
	if (len != 2) {
		Tcl_AppendResult(interp, "Cursor must be a two-element list.", NULL);
		return TCL_ERROR;
	}
	if ( Tcl_ListObjIndex(interp,objv[1],0,&x) != TCL_OK )
	        return TCL_ERROR;
	if ( Tcl_ListObjIndex(interp,objv[1],1,&y) != TCL_OK )
	        return TCL_ERROR;
	if ( Tcl_GetIntFromObj(interp,x,&point.x) != TCL_OK )
	        return TCL_ERROR;
	if ( Tcl_GetIntFromObj(interp,y,&point.y) != TCL_OK )
	        return TCL_ERROR;

	if (ClientToScreen(GetForegroundWindow(), &point) == 0) {
		Tcl_AppendResult(interp, "Cannot convert from client to screen.", NULL);
		return TCL_ERROR;
	}
	if (SetCursorPos(point.x, point.y) == 0) {
		Tcl_AppendResult(interp, "Cannot set cursor position.", NULL);
		return TCL_ERROR;
	}

	return TCL_OK;
}

//
// Getwrect: return coordinates of active window
//
static int
Cwind_GetWindowObjCmd(
	ClientData clientData,
	Tcl_Interp *interp,
	int objc,
	Tcl_Obj *CONST objv[]
)
{
	RECT rect;
	Tcl_Obj *ov[4];
	
	if ( GetWindowRect(GetForegroundWindow(),&rect) == 0 ) {
		Tcl_AppendResult(interp, "Cannot get window coordinates.", NULL);
		return TCL_ERROR;
	}
	ov[0] = Tcl_NewIntObj(rect.left);
	ov[1] = Tcl_NewIntObj(rect.top);
	ov[2] = Tcl_NewIntObj(rect.right);
	ov[3] = Tcl_NewIntObj(rect.bottom);
	Tcl_SetObjResult(interp, Tcl_NewListObj(4,ov) );

	return TCL_OK;
}

//
// Getrect: return coordinates of active window's client area
//
static int
Cwind_GetClientObjCmd(
	ClientData clientData,
	Tcl_Interp *interp,
	int objc,
	Tcl_Obj *CONST objv[]
)
{
	RECT rect;
	Tcl_Obj *ov[4];
	
	if ( GetClientRect(GetForegroundWindow(),&rect) == 0 ) {
		Tcl_AppendResult(interp, "Cannot get window's client coordinates.", NULL);
		return TCL_ERROR;
	}
	ov[0] = Tcl_NewIntObj(rect.left);
	ov[1] = Tcl_NewIntObj(rect.top);
	ov[2] = Tcl_NewIntObj(rect.right);
	ov[3] = Tcl_NewIntObj(rect.bottom);
	Tcl_SetObjResult(interp, Tcl_NewListObj(4,ov) );

	return TCL_OK;
}

//
// click - Send a left|middle|right mouse click to the active window.
//
// According to the microsoft documentation: "The mouse_event() function
// has been superseded. Use SendInput instead." Since I implemented this
// while testing sending mouse clicks to browsers and it works, keep this
// rather than the older mouse_event() calls.
//
static int
click( int button )
{
	INPUT event;

	event.type = INPUT_MOUSE;
	event.mi.dx =
	event.mi.dy = 
	event.mi.mouseData = 0;
	event.mi.dwExtraInfo = GetMessageExtraInfo();

	event.mi.dwFlags = downup[button][down];
	event.mi.time = 0;

	if ( SendInput(1,&event,sizeof(event)) != 1 )
	  return TCL_ERROR;

	Sleep( 50 );

	event.mi.dwFlags = downup[button][up];
	event.mi.time = 0;

	if ( SendInput(1,&event,sizeof(event)) != 1 )
	  return TCL_ERROR;

	return TCL_OK;
}

static int
Cwind_LclickObjCmd( ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[] )
{
  	return click( left );
}

static int
Cwind_MclickObjCmd( ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[] )
{
  	return click( middle );
}

static int
Cwind_RclickObjCmd( ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[] )
{
  	return click( right );
}

//
// Simulate a mouse drag relative to screen coordinates
//
// According to the microsoft documentation: "The mouse_event() function
// has been superseded. Use SendInput instead." and I found that it seemed
// to fail during the click-move-release sequence specifically for the move.
// Using SetCursorPos() and breaking the drag into a series of smaller
// movements rather than on large drag seems to make it more reliable.
//
// I noticed that AutoHotKey uses SendInput. If this routine continues to
// be unreliable, I will have to look into switching to SendInput.
//
static int
drag( Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], enum mbutton button)
{
	LPARAM extra;
	POINT pointTo, pointFrom;
	int len, incX, incY, countX, countY, count, n;
	Tcl_Obj *x, *y;

	if (objc != 2) {
		Tcl_WrongNumArgs(interp, 1, objv, "{x y}");
		return TCL_ERROR;
	}
	if ( Tcl_ListObjLength(interp,objv[1],&len) != TCL_OK )
	        return TCL_ERROR;
	if (len != 2) {
		Tcl_AppendResult(interp, "Destination must be a two-element list.", NULL);
		return TCL_ERROR;
	}
	if ( Tcl_ListObjIndex(interp,objv[1],0,&x) != TCL_OK )
	        return TCL_ERROR;
	if ( Tcl_ListObjIndex(interp,objv[1],1,&y) != TCL_OK )
	        return TCL_ERROR;
	if ( Tcl_GetIntFromObj(interp,x,&pointTo.x) != TCL_OK )
	        return TCL_ERROR;
	if ( Tcl_GetIntFromObj(interp,y,&pointTo.y) != TCL_OK )
	        return TCL_ERROR;

	extra = GetMessageExtraInfo();

	// A drag starts at the current cursor position so retrieve that
	// location
	if (GetCursorPos(&pointFrom) == 0) {
		Tcl_AppendResult(interp, "Cannot get cursor position.", NULL);
		return TCL_ERROR;
	}
	// Press the mouse button
	mouse_event(downup[button][down] | MOUSEEVENTF_ABSOLUTE, pointFrom.x, pointFrom.y, 0, extra);

	Sleep( 25 );

	// Compute a series of smaller moves rather than attempting a single
	// large move. Make no less than 2 moves nor more than approximately
	// 100 pixels per move.
	countX = Abs( (int)pointTo.x - (int)pointFrom.x ) / 100;
	countY = Abs( (int)pointTo.y - (int)pointFrom.y ) / 100;

	count = Max ( countX, countY );

	if ( count < 2 )
	  count = 2;

	incX = ( (int)pointTo.x - (int)pointFrom.x ) / count;
	incY = ( (int)pointTo.y - (int)pointFrom.y ) / count;
	  
	// Do one less than count moves
	for (n = 1; n < count; ++n ) {
	  pointFrom.x = (int)pointFrom.x + incX;
	  pointFrom.y = (int)pointFrom.y + incY;
	  if (SetCursorPos(pointFrom.x, pointFrom.y) == 0) {
		Tcl_AppendResult(interp, "Cannot set cursor position.", NULL);
		return TCL_ERROR;
	  }
	  Sleep( 25 );
	}

	// Do the final move
	if (SetCursorPos(pointTo.x, pointTo.y) == 0) {
		Tcl_AppendResult(interp, "Cannot set cursor position.", NULL);
		return TCL_ERROR;
	}

	Sleep( 50 );

	// Release the mouse button
	mouse_event( downup[button][up] | MOUSEEVENTF_ABSOLUTE, pointTo.x, pointTo.y, 0, extra );

	Sleep( 25 );

	return TCL_OK;
}

//
// Drag? {x y}
// Drag the {left,middle,right} mouse to {x y}
//
static int
Cwind_DraglObjCmd( ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
  	return drag( interp, objc, objv, left );
}
static int
Cwind_DragmObjCmd( ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
  	return drag( interp, objc, objv, middle );
}
static int
Cwind_DragrObjCmd( ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
  	return drag( interp, objc, objv, right );
}

// Simulate a mouse drag relative to window's coordinates
static int
wdrag( Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], enum mbutton button)
{
	LPARAM extra;
	POINT pointTo, pointFrom;
	int len, incX, incY, countX, countY, count, n;
	Tcl_Obj *x, *y;
	
	if (objc != 2) {
		Tcl_WrongNumArgs(interp, 1, objv, "{x y}");
		return TCL_ERROR;
	}
	if ( Tcl_ListObjLength(interp,objv[1],&len) != TCL_OK )
	        return TCL_ERROR;
	if (len != 2) {
		Tcl_AppendResult(interp, "Destination must be a two-element list.", NULL);
		return TCL_ERROR;
	}
	if ( Tcl_ListObjIndex(interp,objv[1],0,&x) != TCL_OK )
	        return TCL_ERROR;
	if ( Tcl_ListObjIndex(interp,objv[1],1,&y) != TCL_OK )
	        return TCL_ERROR;
	if ( Tcl_GetIntFromObj(interp,x,&pointTo.x) != TCL_OK )
	        return TCL_ERROR;
	if ( Tcl_GetIntFromObj(interp,y,&pointTo.y) != TCL_OK )
	        return TCL_ERROR;

	if (ClientToScreen(GetForegroundWindow(), &pointTo) == 0) {
		Tcl_AppendResult(interp, "Cannot convert from client to screen.", NULL);
		return TCL_ERROR;
	}

	extra = GetMessageExtraInfo();

	// A drag starts at the current cursor position so retrieve that
	// location
	if (GetCursorPos(&pointFrom) == 0) {
		Tcl_AppendResult(interp, "Cannot get cursor position.", NULL);
		return TCL_ERROR;
	}
	// Press the mouse button
	mouse_event(downup[button][down] | MOUSEEVENTF_ABSOLUTE, pointFrom.x, pointFrom.y, 0, extra);

	Sleep( 25 );
	//
	// Compute a series of smaller moves rather than attempting a single
	// large move. Make no less than 2 moves nor more than approximately
	// 100 pixels per move.
	countX = Abs( (int)pointTo.x - (int)pointFrom.x ) / 100;
	countY = Abs( (int)pointTo.y - (int)pointFrom.y ) / 100;

	count = Max ( countX, countY );

	if ( count < 2 )
	  count = 2;

	incX = ( (int)pointTo.x - (int)pointFrom.x ) / count;
	incY = ( (int)pointTo.y - (int)pointFrom.y ) / count;
	  
	// Do one less than count moves
	for (n = 1; n < count; ++n ) {
	  pointFrom.x = (int)pointFrom.x + incX;
	  pointFrom.y = (int)pointFrom.y + incY;
	  if (SetCursorPos(pointFrom.x, pointFrom.y) == 0) {
		Tcl_AppendResult(interp, "Cannot set cursor position.", NULL);
		return TCL_ERROR;
	  }
	  Sleep( 25 );
	}

	// Do the final move
	if (SetCursorPos(pointTo.x, pointTo.y) == 0) {
		Tcl_AppendResult(interp, "Cannot set cursor position.", NULL);
		return TCL_ERROR;
	}

	Sleep( 50 );

	// Release the mouse button
	mouse_event( downup[button][up] | MOUSEEVENTF_ABSOLUTE, pointTo.x, pointTo.y, 0, extra );

	Sleep( 25 );

	return TCL_OK;
}

//
// Wdrag? {x y}
// Drag the {left,middle,right} mouse to {x y} relative to coordinates of active window
//
static int
Cwind_WdraglObjCmd( ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
  	return wdrag( interp, objc, objv, left );
}
static int
Cwind_WdragmObjCmd( ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
  	return wdrag( interp, objc, objv, middle );
}
static int
Cwind_WdragrObjCmd( ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
  	return wdrag( interp, objc, objv, right );
}

/*
 * For comleteness, include a "version" command
 * Returns "cwind x.x.x (nn-bit) Compiled Mmm dd, yyyy at hh:mm:ss by GNU C vvvvvv"
 */
static int
Cwind_VersionObjCmd( ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
  	Tcl_Obj *version;
	char package[] = PACKAGE;
	char *date = __DATE__; /* MMM NN NNNN */

	package[0] = toupper( package[0] );

	version = Tcl_NewStringObj( package, -1 );
	Tcl_AppendToObj( version, " " VERSION, -1 );
#ifdef __WIN64__
	Tcl_AppendToObj( version, " (64-bit)", 9 );
#else
	Tcl_AppendToObj( version, " (32-bit)", 9 );
#endif
	Tcl_AppendToObj( version, " Compiled ", 10 );

	Tcl_AppendToObj( version, date, 4 );
	if ( date[4] == ' ' )
	  Tcl_AppendToObj( version, date+5, 1 );
	else
	  Tcl_AppendToObj( version, date+4, 2 );
	Tcl_AppendToObj( version, ",", 1 );
	Tcl_AppendToObj( version, date+6, 5 );

	Tcl_AppendToObj( version, " at " __TIME__, 12 );
#ifdef __GNUC__
	Tcl_AppendToObj( version, " by GNU C " __VERSION__, -1 );
#endif

	Tcl_SetObjResult( interp, version );

	return TCL_OK;
}

//
// Create each command as <nameSpace>::<command> and export it to allow
// one to do a "namespace import <nameSpace>::<command>"
//
Tcl_Command
AddCommand(
	Tcl_Interp *interp,
	char *nameSpace,
	char *cmdName,
	Tcl_ObjCmdProc *cmdProc
)
{
	char *cmdPath;
	Tcl_DString dString;
	Tcl_Command cmdToken;

	Tcl_DStringInit(&dString);

	// Create Namespace::Command
	if (nameSpace != NULL) {
		Tcl_DStringAppend(&dString, nameSpace, -1);
	}
	Tcl_DStringAppend(&dString, "::", -1);
	Tcl_DStringAppend(&dString, cmdName, -1);

	cmdPath = Tcl_DStringValue(&dString);
	cmdToken = Tcl_FindCommand(interp, cmdPath, (Tcl_Namespace *)NULL, 0);
	if (cmdToken != NULL) {
		Tcl_DStringFree(&dString);
		return cmdToken;
	}
	cmdToken = Tcl_CreateObjCommand(interp, cmdPath, cmdProc, NULL, NULL);
	Tcl_DStringFree(&dString);

	Tcl_Namespace *nsPtr;
	int dontResetList = 0;

	// Namespace export
	nsPtr = Tcl_FindNamespace(interp, nameSpace, (Tcl_Namespace *)NULL, TCL_LEAVE_ERR_MSG);
	if (nsPtr == NULL) {
		return NULL;
	}
	if (Tcl_Export(interp, nsPtr, cmdName, dontResetList) != TCL_OK) {
		return NULL;
	}
	return cmdToken;
}

//
// DLL initialize entry point
//
int
Cwind_Init(
	Tcl_Interp *interp
)
{
	Tcl_Namespace *spacePtr;

#ifdef USE_TCL_STUBS
	if (Tcl_InitStubs(interp, "8.1", 0) == NULL)
		return TCL_ERROR;
#endif // USE_TCL_STUBS

	// Namespace creation
	spacePtr = Tcl_CreateNamespace(interp, NAMESPACE, (ClientData)0, (Tcl_NamespaceDeleteProc *) NULL);
	if (spacePtr == NULL)
		return TCL_ERROR;

	// Create the commands
	AddCommand(interp, NAMESPACE, "gettext", Cwind_GetTextObjCmd);
	AddCommand(interp, NAMESPACE, "send", Cwind_SendObjCmd);
	AddCommand(interp, NAMESPACE, "show", Cwind_ShowObjCmd);
	AddCommand(interp, NAMESPACE, "getcb", Cwind_GetCbObjCmd);
	AddCommand(interp, NAMESPACE, "putcb", Cwind_PutCbObjCmd);
	AddCommand(interp, NAMESPACE, "wlist", Cwind_WlistObjCmd);
	AddCommand(interp, NAMESPACE, "sleeptime", Cwind_SleeptimeObjCmd);
	AddCommand(interp, NAMESPACE, "restore", Cwind_RestoreObjCmd);
	AddCommand(interp, NAMESPACE, "minimize", Cwind_MinimizeObjCmd);
	AddCommand(interp, NAMESPACE, "maximize", Cwind_MaximizeObjCmd);
	AddCommand(interp, NAMESPACE, "state", Cwind_StateObjCmd);
	AddCommand(interp, NAMESPACE, "bottom", Cwind_BottomObjCmd);
	AddCommand(interp, NAMESPACE, "top", Cwind_TopObjCmd);
	AddCommand(interp, NAMESPACE, "topmost", Cwind_TopMostObjCmd);
	AddCommand(interp, NAMESPACE, "notopmost", Cwind_NoTopMostObjCmd);
	AddCommand(interp, NAMESPACE, "sendmessage", Cwind_SendmessageObjCmd);
	AddCommand(interp, NAMESPACE, "postmessage", Cwind_PostmessageObjCmd);
	AddCommand(interp, NAMESPACE, "getpos", Cwind_GetposObjCmd);
	AddCommand(interp, NAMESPACE, "setpos", Cwind_SetposObjCmd);
	AddCommand(interp, NAMESPACE, "getwpos", Cwind_GetwposObjCmd);
	AddCommand(interp, NAMESPACE, "setwpos", Cwind_SetwposObjCmd);
	AddCommand(interp, NAMESPACE, "lclick", Cwind_LclickObjCmd);
	AddCommand(interp, NAMESPACE, "mclick", Cwind_MclickObjCmd);
	AddCommand(interp, NAMESPACE, "rclick", Cwind_RclickObjCmd);

	// New for version 1.5
	AddCommand(interp, NAMESPACE, "getwrect", Cwind_GetWindowObjCmd);
	AddCommand(interp, NAMESPACE, "getrect", Cwind_GetClientObjCmd);
	AddCommand(interp, NAMESPACE, "dragl", Cwind_DraglObjCmd);
	AddCommand(interp, NAMESPACE, "dragm", Cwind_DragmObjCmd);
	AddCommand(interp, NAMESPACE, "dragr", Cwind_DragrObjCmd);
	AddCommand(interp, NAMESPACE, "wdragl", Cwind_WdraglObjCmd);
	AddCommand(interp, NAMESPACE, "wdragm", Cwind_WdragmObjCmd);
	AddCommand(interp, NAMESPACE, "wdragr", Cwind_WdragrObjCmd);

	AddCommand(interp, NAMESPACE, "version", Cwind_VersionObjCmd);

	return Tcl_PkgProvide(interp, PACKAGE, VERSION);
}

#if 0
//
// Allowing a Tcl script to activate and send arbitrary keystrokes to any
// application is clearly not "safe"
//
int
Cwind_SafeInit(
	Tcl_Interp *interp
)
{
  	return Cwind_Init( interp );
}
#endif
